from scipy.stats import binom as _binom, poisson as _poisson, geom as _geom, expon as _expon
from typing import Union, Any, Iterable, Sequence
import random as _rd
import numpy as _np
import abc as _abc

from coota.sets import *


class Chooser(object):
    @_abc.abstractmethod
    def choice(self, x: Sequence) -> Any:
        """
        This method specifies how the chooser selects an item.
        :param x: The data source from which the chooser chooses.
        :return: A single item chosen from `x`.
        """
        pass

    @_abc.abstractmethod
    def choices(self, x: Sequence, n: int) -> Sequence:
        """
        This method specifies how the chooser selects a batch of items.
        :param x: The data source from which the chooser chooses.
        :param n: The number of choices.
        :return: A list of items chosen from `x`.
        """
        pass


class DefaultChooser(Chooser):
    def choice(self, x: Sequence) -> Any:
        return _rd.choice(x)

    def choices(self, x: Sequence, n: int) -> Sequence:
        return _rd.choices(x, k=n)


class GaussianChooser(Chooser):
    @staticmethod
    def _analyse(x: Sequence) -> [float, float, int]:
        s = 0.5 * len(x)
        return s, s / 3

    def choices(self, x: Sequence, n: int) -> Sequence:
        choices = _np.random.normal(*self._analyse(x), size=n)
        r = []
        for c in choices:
            r.append(x[int(c)])
        return r

    def choice(self, x: Sequence) -> Any:
        return x[int(_np.random.normal(*self._analyse(x), size=1))]


class DiscreteChooser(Chooser):
    @_abc.abstractmethod
    def get_weights(self, length: int) -> list[float]:
        pass

    def choice(self, x: Sequence) -> Any:
        return _rd.choices(population=x, weights=self.get_weights(len(x)), k=1)[0]

    def choices(self, x: Sequence, n: int) -> Sequence:
        return _rd.choices(population=x, weights=self.get_weights(len(x)), k=n)


class BinomialChooser(DiscreteChooser):
    def __init__(self, p: float):
        self.p: float = p
        self._weights: Union[list[float], None] = None

    def get_weights(self, length: int) -> list[float]:
        if self._weights is None:
            self._weights = _binom.pmf(_np.arange(0, length, 1), length, self.p)
        return self._weights


class PoissonChooser(DiscreteChooser):
    def __init__(self, lam: float):
        self.lam: float = lam
        self._weights: Union[list[float], None] = None

    def get_weights(self, length: int) -> list[float]:
        if self._weights is None:
            self._weights = _poisson.pmf(_np.arange(0, length, 1), self.lam)
        return self._weights


class GeomChooser(DiscreteChooser):
    def __init__(self, k: float, p: float):
        self.k: float = k
        self.p: float = p
        self._weights: Union[list[float], None] = None

    def get_weights(self, length: int) -> list[float]:
        if self._weights is None:
            self._weights = _geom.pmf(self.k, self.p)
        return self._weights


class ExponChooser(DiscreteChooser):
    def __init__(self, lam: float):
        self.lam: float = lam
        self._weights: Union[list[float], None] = None

    def get_weights(self, length: int) -> list[float]:
        if self._weights is None:
            self._weights = _expon.pmf(_np.arange(0, length, 1), 1 / self.lam)
        return self._weights


class Association(object):
    def __init__(self, the_other_generator: Any):
        """
        :param the_other_generator: The other generator associated with this generator.
            That generator must generate before this generator.
        """
        if not isinstance(the_other_generator, Generator):
            raise TypeError("'obj' must be a Generator.")
        self._other_generator = the_other_generator

    def get_the_other_generator(self) -> Any:
        """
        :return: The other generator.
        """
        return self._other_generator

    @_abc.abstractmethod
    def associate(self, g: Any, the_other_generator_output: Any) -> Any:
        """
        This method specifies the association between the output of two generators.
        :param g: The generator whose `set_association()` is called with the association given to.
        :param the_other_generator_output: The output generated by the other generator.
        :return: Anything. If not none, the return will be returned by `generate()`, or the process will continue.
        """
        pass


class TheSameAs(Association):
    def associate(self, g: Any, the_other_generator_output: Any) -> Any:
        return the_other_generator_output


class GeneratorOutput(object):
    def __init__(self, output: Any):
        """
        :param output: The output.
        """
        self._output = output

    def get_output(self) -> Any:
        """
        :return: The output.
        """
        return self._output


class Generator(object):
    def __init__(self, *default_args, **args):
        """
        :param default_args: Given to `make()` when no arguments are given to `generate()`.
            For example, in a **GeneratorSequence**, `generate()` is called with no arguments,
            then the `default_args` will be given to `make()`.
            Required arguments are listed in the specific generators.
        :param args: Global arguments for the generator. Required arguments are listed in the specific generators.
        """
        self._parseable: bool = True
        self._uc: bool = True
        self._args: dict[str: Any] = {} if args is None else args
        self._default_args: tuple = default_args
        self._chooser: Chooser = DefaultChooser()
        self._source_cache: Union[Sequence, None] = None
        self._last: Any = None
        self._association: Union[Association, None] = None

    def __str__(self):
        return f"[Generator({self.get_args()})]"

    def _get_source_cache(self) -> Union[Sequence, None]:
        """
        :return: The cache of the source.
        """
        return self._source_cache

    def _set_source_cache(self, source_cache: Sequence) -> None:
        """
        :param source_cache: The cache of the source.
        """
        self._source_cache = source_cache

    def _get_last(self) -> Any:
        """
        :return: Last generated data.
        """
        return self._last

    def _set_last(self, last: Any) -> None:
        """
        :param last: Last generated data.
        :return:
        """
        self._last = last

    def get_parseable(self) -> bool:
        """
        :return: Whether the generator can be parsed as an argument.
        """
        return self._parseable

    def set_parseable(self, parseable: bool) -> None:
        """
        :param parseable: Whether the generator can be parsed as an argument. If false,
            the generator will be recognized as an argument itself instead of being parsed into an actual output.
        :return:
        """
        self._parseable = parseable

    def get_uc(self) -> bool:
        """
        :return: Whether the generator uses the cache of the source.
        """
        return self._uc

    def set_uc(self, use_cache: bool) -> None:
        """
        :param use_cache: Whether the generator uses the cache of the source. If true,
            the generator will only call `source()` once and use the cache instead after that.
            It's true by default. Set it to false if your source is not always static.
        :return:
        """
        self._uc = use_cache

    def get_args(self) -> dict:
        """
        :return: The global arguments of the generator.
        """
        return self._args

    def get_arg(self, name: str, required_type: type = object) -> Any:
        """
        :param name: The argument's name.
        :param required_type: The type of argument you require. If any type of argument is acceptable,
            set it to `object` which is also by default.
        :return: The argument's value. It can be `None` if the argument does not exist
            or is not of the same type as required.
        """
        args = self.get_args()
        if name not in args:
            return None
        arg = args[name]
        return arg if isinstance(arg, required_type) else None

    def get_arg_or_default(self, name: str, default: Any, required_type: type = object) -> Any:
        """
        :param name: The argument's name.
        :param default: The default value.
        :param required_type: The type of argument you require. If any type of argument is acceptable,
            set it to `object` which is also by default.
        :return: The argument's value. The default value will be returned if the argument does not exist
            or is not of the same type as required.
        """
        arg = self.get_arg(name, required_type)
        return default if arg is None else arg

    def get_required_arg(self, name: str, required_type: type = object) -> Any:
        """
        :param name: The argument's name.
        :param required_type: The type of argument you require. If any type of argument is acceptable,
            set it to `object` which is also by default.
        :return: The argument's value.
        :exception AttributeError: The argument does not exist.
        :exception TypeError: The argument is not of the same type as required.
        """
        arg = self.get_arg(name)
        if arg is None:
            raise AttributeError(f"'Args' must contain keys '{name}'.")
        if not isinstance(arg, required_type):
            raise TypeError(f"Expecting '{name}' to be type of {required_type}, but got {type(arg)} instead.")
        return arg

    def get_default_args(self) -> tuple:
        """
        :return: The default arguments to be given to `make()`.
        """
        return self._default_args

    def get_chooser(self) -> Chooser:
        """
        :return: The generator's chooser.
        """
        return self._chooser

    def set_chooser(self, chooser: Chooser) -> None:
        """
        :param chooser: Set the generator's chooser. A **DefaultChooser** is used by default.
            If you want to change the behavior of choosing, such as making it fit a certain distribution,
            you can do so by changing the chooser object.
        :return:
        """
        self._chooser = chooser

    def get_association(self) -> Union[Association, None]:
        """
        :return: The association with the other generator.
        """
        return self._association

    def set_association(self, association: Association) -> None:
        """
        :param association: The association with the other generator.
        :return:
        """
        self._association = association

    def choice(self) -> Any:
        """
        :return: One single item chosen from the source by the chooser.
        """
        return self.get_chooser().choice(self.get_source())

    def choices(self, n: int) -> Sequence:
        """
        :param n: The number of items.
        :return: A batch of items chosen from the source by the chooser.
        """
        s = self.get_source()
        return self.get_chooser().choices(s, n)

    @_abc.abstractmethod
    def source(self) -> Sequence:
        """
        This method specifies what data the generator may generate.
        :return: The original dataset from which the generator selects.
        """
        pass

    def get_source(self) -> Sequence:
        """
        :return: The same as `source()` returns. If `use_cache` is true, returns the source cache instead.
        """
        cache_on = self.get_uc()
        sc = self._get_source_cache()
        if cache_on and sc is None:
            self._set_source_cache(self.source())
        return self._get_source_cache() if cache_on else self.source()

    @_abc.abstractmethod
    def make(self, *args) -> Any:
        """
        This method specifies how to generate data.
        :param args: Optional arguments.
        :return: Anything.
        """
        return self.choice()

    def generate(self, *args, parse: bool = True) -> Any:
        """
        :param args: Optional arguments given to `make()`.
        :param parse: Whether to resolve the generator in parameters and return values. True: return an output.
            False: return the generator itself.
        :return: Anything.
        :exception LookupError: The associated generator generated before the generator to which is associated has
            generated.
        """
        r = None
        try:
            if not parse or not self.get_parseable():
                r = self
                return r
            a = self.get_association()
            if a is not None:
                og: Generator = a.get_the_other_generator()
                ogl = og._get_last()
                if ogl is None:
                    raise LookupError("The associated generator must not generate until the generator to which is "
                                      "associated has generated.")
                r = a.associate(self, ogl)
                if r is not None:
                    return r
            if args == ():
                args = self.get_default_args()
            args = list(args)
            for i in range(len(args)):
                arg = args[i]
                if isinstance(arg, Generator):
                    args[i] = arg.generate(parse=parse)
            r = self.make(*args)
            if isinstance(r, Generator):
                r = r.generate(parse=parse)
            return r
        finally:
            self._set_last(r)

    def output(self, *args, parse: bool = False) -> GeneratorOutput:
        """
        :param args: Optional arguments given to `make()`.
        :param parse: Whether to resolve the generator in parameters and return values.
        :return: The return value of `generate()` wrapped as a **GeneratorOutput**.
        """
        return GeneratorOutput(self.generate(*args, parse=parse))


class ItertableGenerator(Generator):
    def __init__(self, *default_args, **args):
        super().__init__(*default_args, **args)
        self._pointer: Any = 0
        self.initialize()

    def __iter__(self) -> Iterable:
        self.initialize()
        return self

    def __next__(self) -> Any:
        return self.generate()

    def get_pointer(self) -> Any:
        """
        :return: The pointer.
        """
        return self._pointer

    def set_pointer(self, pointer: Any) -> None:
        """
        :param pointer: The pointer which is 0 in default.
        :return:
        """
        self._pointer = pointer

    def choice(self) -> Any:
        """
        :return: `self.get_source()[pointer]`. The pointer must be an integer when calling this method.
        """
        p = self.get_pointer()
        if not isinstance(p, int):
            raise TypeError("When this method is called, the pointer must be of type int.")
        return self.get_source()[p]

    def choices(self, n: int) -> Sequence:
        """
        :param n: the number
        :return: `self.get_source()[pointer: pointer + n]`. The pointer must be an integer when calling this method.
        """
        p = self.get_pointer()
        if not isinstance(p, int):
            raise TypeError("When this method is called, the pointer must be of type int.")
        return self.get_source()[p: p + n]

    @_abc.abstractmethod
    def initialize(self) -> None:
        """
        This is a callback used to set initialization operations such as pointers.
        :return:
        """
        pass

    @_abc.abstractmethod
    def step(self) -> bool:
        """
        This method specifies how the iterator iterates.
        :return: True: continue iteration. False: stop iteration.
        """
        return False

    def generate(self, *args, parse: bool = True) -> Any:
        try:
            return super(ItertableGenerator, self).generate(*args, parse=parse)
        finally:
            if not self.step():
                raise StopIteration


class LetterGenerator(Generator):
    """
    Generate a letter.
    """

    def source(self) -> Sequence:
        return LETTER_SET

    def make(self, *args) -> Any:
        return self.choice()


class StringGenerator(LetterGenerator):
    """
    Generate a string of letters.
    Arguments:
        (Required) length of string
    """

    def make(self, *args) -> Any:
        return "".join(self.choices(args[0]))


class NumberGenerator(StringGenerator):
    """
    Generate a string of numbers.
    Arguments:
        (Required) length of string
    """

    def source(self) -> Sequence:
        return NUMBER_SET

    def make(self, *args) -> Any:
        s = super(NumberGenerator, self).make(*args)
        if len(s) > 1 and s[0] == "0":
            s[0] = self.get_chooser().choice("123456789")
        return s


class LetterAndNumberGenerator(StringGenerator):
    """
    Generate a string of letters and numbers.
    Arguments:
        (Required) length of string
    """

    def source(self) -> Sequence:
        return NUMBER_SET + LETTER_SET


class IntGenerator(Generator):
    """
    Generate an integer x(a ≤ x < b).
    Global Arguments:
        (Required) start: a
        (Required) stop: b
    """

    def source(self) -> Sequence:
        return range(self.get_required_arg("start", required_type=int),
                     self.get_required_arg("stop", required_type=int))

    def make(self, *args) -> Any:
        return self.choice()


class IntIterable(ItertableGenerator):
    def source(self) -> Sequence:
        return self.get_pointer()

    def make(self, *args) -> Any:
        return self.get_source()

    def initialize(self) -> None:
        self.set_uc(False)
        self.set_pointer(self.get_required_arg("start", required_type=int))

    def step(self) -> bool:
        self._pointer += self.get_arg_or_default("step", 1, required_type=int)
        return self._pointer <= self.get_required_arg("stop", required_type=int)



from typing import Any, Sequence
import datetime as _dt


from coota import generator as _g, toolkit as _tk
from coota.preset.sets import *


class TimeGenerator(_g.IntGenerator):
    def __init__(self, *default_args, **args):
        if "start" not in args:
            args["start"] = int(_dt.datetime.now().timestamp())
        elif isinstance(args["start"], str):
            args["start"] = int(_tk.parse_time(args["start"]).timestamp())
        elif isinstance(args["start"], _dt.datetime):
            args["start"] = int(args["start"].timestamp())
        if "stop" not in args:
            args["stop"] = int((_dt.datetime.now() + _dt.timedelta(days=365)).timestamp())
        elif isinstance(args["stop"], str):
            args["stop"] = int(_tk.parse_time(args["stop"]).timestamp())
        elif isinstance(args["stop"], _dt.datetime):
            args["stop"] = int(args["stop"].timestamp())
        super(TimeGenerator, self).__init__(*default_args, **args)

    def make(self, *args) -> Any:
        return _dt.datetime.utcfromtimestamp(self.choice())


class EmailGenerator(_g.StringGenerator):
    def source(self) -> Sequence:
        return _g.NUMBER_SET + _g.LETTER_SET + "."

    def make(self, *args) -> Any:
        domain = self.get_required_arg("domain")
        return f"%s@{domain}" % "".join(self.choices(args[0]))


class QQMailGenerator(EmailGenerator):
    def __init__(self, *default_args, **args):
        args["domain"] = "qq.com"
        super().__init__(*default_args, **args)

    def source(self) -> Sequence:
        return _g.NUMBER_SET


class NameGenerator(_g.Generator):
    def source(self) -> Sequence:
        return NAME_MALE + NAME_FEMALE

    def make(self, *args) -> Any:
        return self.choice()



from functools import singledispatch as _singledispatch
from matplotlib import pyplot as _plt


from coota import generator as _g, generator_sequence as _gs


@_singledispatch
def show(obj) -> None:
    raise TypeError(f"No known case for type {type(obj)}.")


@show.register(_g.Chooser)
def _(obj: _g.Chooser, title: str = "") -> None:
    d = obj.choices(range(100), 100)
    _plt.hist(d)
    _plt.xlabel("Item")
    _plt.ylabel("Possibility")
    _plt.title(title)
    _plt.show()


@show.register(_g.Generator)
def _(obj: _g.Generator) -> None:
    print("[%s] {" % str(type(obj))[8:-2])
    print(f"....Arguments -> {obj.get_args()}")
    print(f"....Default Arguments -> {obj.get_default_args()}")
    print(f"....Parseable -> {obj.get_parseable()}")
    print(f"....Use Cache -> {obj.get_uc()}")
    print("}")


@show.register(_g.GeneratorOutput)
def _(obj: _g.GeneratorOutput) -> None:
    show(obj.get_output())


@show.register(_gs.GeneratorSequence)
def _(obj: _gs.GeneratorSequence) -> None:
    print(obj)


@show.register(list)
def _(obj: list, title: str = "") -> None:
    _plt.hist(obj)
    _plt.xlabel("Item")
    _plt.ylabel("Count")
    _plt.title(title)
    _plt.show()


@show.register(str)
def _(obj: str) -> None:
    print(obj)



NAME_MALE: tuple = (
    "Aaron", "Abel", "Abraham", "Adam", "Adrian", "Aidan", "Alva", "Alex", "Alexander", "Alan", "Albert",
    "Alfred", "Andrew", "Andy", "Angus", "Anthony", "Apollo", "Arnold", "Arthur", "August", "Austin",
    "Ben", "Benjamin", "Bert", "Benson", "Bill", "Billy", "Blake", "Bob", "Bobby", "Brad", "Brandon",
    "Brant", "Brent", "Brian", "Bryan", "Brown", "Bruce", "Caleb", "Cameron", "Carl", "Carlos", "Cary",
    "Caspar", "Caspar", "Charles", "Cheney", "Chris", "Christian", "Christopher", "Clark", "Cliff", "Cody",
    "Cole", "Colin", "Cosmo", "Daniel", "Denny", "Darwin", "David", "Dennis", "Derek", "Dick", "Donald",
    "Douglas", "Duke", "Dylan", "Eddie", "Edgar", "Edison", "Edmund", "Edward", "Edwin", "Elijah",
    "Elliott", "Elvis", "Eric", "Ethan", "Eugene", "Evan", "Ford", "Francis", "Frank", "Franklin", "Fred",
    "Gabriel", "Gaby", "Garfield", "Gary", "Gavin", "Geoffrey", "George", "Gino", "Glen", "Glendon",
    "Hank", "Hardy", "Harrison", "Harry", "Hayden", "Henry", "Hilton", "Hugo", "Hunk", "Howard", "Henry",
    "Ian", "Ivan", "Isaac", "Isaiah", "Jack", "Jackson", "Jacob", "James", "Jason", "Jay", "Jeffery",
    "Jerome", "Jerry", "Jesse", "Jim", "Jimmy", "Joe", "John", "Johnny", "Jonathan", "Jordan", "Jose",
    "Joseph", "Joshua", "Justin", "Keith", "Ken", "Kennedy", "Kenneth", "Kenny", "Kevin", "Kyle", "Lance",
    "Larry", "Laurent", "Lawrence", "Leander", "Lee", "Leo", "Leonard", "Leopold", "Leslie", "Loren",
    "Lori", "Lorin", "Louis", "Luke", "Marcus", "Marcy", "Mark", "Marks", "Mars", "Marshal", "Martin",
    "Marvin", "Mason", "Matthew", "Max", "Michael", "Mickey", "Mike", "Nathan", "Nathaniel", "Neil",
    "Nelson", "Nicholas", "Nick", "Noah", "Norman", "Oliver", "Oscar", "Owen", "Patrick", "Paul", "Peter",
    "Philip", "Phoebe", "Quentin", "Randall", "Randolph", "Randy", "Ray", "Raymond", "Reed", "Rex",
    "Richard", "Richie", "Riley", "Robert", "Robin", "Robinson", "Rock", "Roger", "Ronald", "Rowan", "Roy",
    "Ryan", "Sam", "Sammy", "Samuel", "Scott", "Sean", "Shawn", "Sidney", "Simon", "Solomon", "Spark",
    "Spencer", "Spike", "Stanley", "Steve", "Steven", "Stewart", "Stuart", "Taylor", "Terence", "Terry",
    "Ted", "Tim", "Timothy", "Todd", "Tommy", "Tom", "Thomas", "Tony", "Tyler", "Ulysses", "Van", "Vern",
    "Vernon", "Victor", "Vincent", "Warner", "Warren", "Wayne", "Wesley", "William", "Willy", "Zack",
    "Zachary",
)
NAME_FEMALE: tuple = (
    "Abigail", "Abby", "Ada", "Adelaide", "Adeline", "Alexandra", "Ailsa", "Aimee", "Alexis", "Alice",
    "Alicia", "Alina", "Allison", "Alyssa", "Amanda", "Amy", "Amber", "Anastasia", "Andrea", "Angel",
    "Angela", "Angelia", "Angelina", "Ann", "Anna", "Anne", "Annie", "Anita", "Ariel", "April", "Ashley",
    "Audrey", "Aviva", "Barbara", "Barbie", "Beata", "Beatrice", "Becky", "Bella", "Bess", "Bette",
    "Betty", "Blanche", "Bonnie", "Brenda", "Brianna", "Britney", "Brittany", "Camille", "Candice",
    "Candy", "Carina", "Carmen", "Carol", "Caroline", "Carry", "Carrie", "Cassandra", "Cassie",
    "Catherine", "Cathy", "Chelsea", "Charlene", "Charlotte", "Cherry", "Cheryl", "Chloe", "Chris",
    "Christina", "Christine", "Christy", "Cindy", "Claire", "Clair", "Claudia", "Clement", "Cloris",
    "Connie", "Constance", "Cora", "Corrine", "Crystal", "Daisy", "Daphne", "Darcy", "Dave", "Debbie",
    "Deborah", "Debra", "Demi", "Diana", "Dolores", "Donna", "Dora", "Doris", "Edith", "Editha", "Elaine",
    "Eleanor", "Elizabeth", "Ella", "Ellen", "Ellie", "Emerald", "Emily", "Emma", "Enid", "Elsa", "Erica",
    "Estelle", "Esther", "Eudora", "Eva", "Eve", "Evelyn", "Fannie", "Fay", "Fiona", "Flora", "Florence",
    "Frances", "Frederica", "Frieda", "Gina", "Gillian", "Gladys", "Gloria", "Grace", "Greta", "Gwendolyn",
    "Hannah", "Haley", "Hebe", "Helena", "Hellen", "Henna", "Heidi", "Hillary", "Ingrid", "Isabella",
    "Irene", "Iris", "Ivy", "Jacqueline", "Jade", "Jamie", "Jane", "Janet", "Jasmine", "Jean", "Jenna",
    "Jennifer", "Jenny", "Jessica", "Jessie", "Jill", "Joan", "Joanna", "Jocelyn", "Joliet", "Josephine",
    "Josie", "Joy", "Joyce", "Judith", "Judy", "Julia", "Juliana", "Julie", "June", "Karen", "Katherine",
    "Kate", "Kathy", "Katie", "Kate", "Katy", "Katrina", "Kay", "Kayla", "Kelly", "Kelsey", "Kimberly",
    "Kitty", "Lassie", "Laura", "Lauren", "Lena", "Lydia", "Lillian", "Lily", "Linda", "Lindsay", "Lisa",
    "Liz", "Lora", "Laura", "Lorraine", "Louisa", "Louise", "Lucia", "Lucy", "Lucine", "Lulu", "Lydia",
    "Lynn", "Mabel", "Madeline", "Maggie", "Mamie", "Manda", "Mandy", "Margaret", "Mariah", "Marilyn",
    "Martha", "Mavis", "Mary", "Matilda", "Maureen", "Mavis", "Maxine", "May", "Mayme", "Mayme", "Melinda",
    "Melissa", "Melody", "Mercedes", "Meredith", "Mia", "Michelle", "Milly", "Miranda", "Miriam", "Miya",
    "Molly", "Monica", "Morgan", "Nancy", "Natalie", "Natasha", "Nicole", "Nikita", "Nina", "Nora",
    "Norma", "Nydia", "Octavia", "Olivia", "Ophelia", "Oprah", "Pamela", "Patricia", "Patty", "Paula",
    "Pauline", "Pearl", "Peggy", "Philomena", "Phoebe", "Phyllis", "Polly", "Priscilla", "Rachel",
    "Rebecca", "Regina", "Rita", "Rose", "Roxanne", "Ruth", "Sabrina", "Sally", "Sandra", "Samantha",
    "Sami", "Sandra", "Sandy", "Sarah", "Sara", "Savannah", "Scarlett", "Selma", "Selina", "Serena",
    "Sharon", "Sheila", "Shelley", "Shelly", "Sherry", "Shirley", "Sierra", "Silvia", "Sonia", "Sophia",
    "Stacy", "Stella", "Stephanie", "Sue", "Sunny", "Susan", "Tamara", "Tammy", "Tanya", "Tasha", "Teresa",
    "Tess", "Tiffany", "Tina", "Tonya", "Tracy", "Ursula", "Vanessa", "Venus", "Vera", "Vicky",
    "Victoria", "Violet", "Virginia", "Vita", "Vivian", "Wanda", "Wendy", "Whitney", "Wynne", "Winnie",
    "Yolanda", "Yvette", "Yvonne", "Zara", "Zelda", "Zoey", "Zoe", "Zora"
)



class Configuration(object):
    def __init__(self):
        self.charset: str = "utf-8"
        self.number_set: str = "1234567890"
        self.alpha_set: str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.use_source_cache: bool = False



from typing import Sequence
import datetime as _dt
import re as _re


UNITS = ("year", "month", "day", "hour", "minute", "second", "microsecond")

UNITS_FEATURES = (
    ("year", lambda t: 3 < t <= 4, lambda t: True),
    ("month", lambda t: 0 < t <= 2, lambda t: 0 < t < 13),
    ("day", lambda t: 0 < t <= 2, lambda t: 0 < t < 32),
    ("hour", lambda t: 0 < t <= 2, lambda t: -1 < t < 25),
    ("minute", lambda t: 0 < t <= 2, lambda t: -1 < t < 61),
    ("second", lambda t: 0 < t <= 2, lambda t: -1 < t < 61),
    ("microsecond", lambda t: 0 < t <= 6, lambda t: -1 < t < 1000000)
)


def parse_time_type(t: str, exclude: Sequence) -> [str, int]:
    length = len(t)
    t = int(t)
    for unit in UNITS_FEATURES:
        if unit[1](length) and unit[0] not in exclude and unit[2](t):
            return unit[0], t
    return "unknown", t


def parse_time(time: str) -> _dt.datetime:
    cases = _re.findall(r"\d+", time)
    exclude = []
    now = _dt.datetime.now()
    matched = [now.year, now.month, now.day, now.hour, now.minute, now.second, now.microsecond]
    for t in cases:
        time_type, t = parse_time_type(t, exclude)
        if time_type == "unknown":
            continue
        exclude.append(time_type)
        matched[UNITS.index(time_type)] = t
    return _dt.datetime(year=matched[0], month=matched[1], day=matched[2], hour=matched[3], minute=matched[4],
                        second=matched[5], microsecond=matched[6])



from typing import Iterable, Union


from coota import generator as _g


class GeneratorSequence(object):
    def __init__(self, *sequence, n: int, t: type = str):
        self._sequential: tuple = sequence
        self._n: int = n
        self._t: type = t
        self._i: int = n

    def __len__(self):
        return len(self.get_sequence())

    def __str__(self):
        s = ""
        for i in self.get_sequence():
            s += str(i)
        return s

    def __iter__(self) -> Iterable:
        self._i = self._n
        return self

    def __next__(self) -> Union[str, tuple, list]:
        if self._i < 1:
            self._i = self._n
            raise StopIteration
        self._i -= 1
        return self._next()

    def _initialize(self) -> None:
        for i in self.get_sequence():
            if isinstance(i, _g.ItertableGenerator):
                i.initialize()

    def _next_str(self) -> str:
        s = ""
        for i in range(len(self)):
            j = self.get_sequence()[i]
            if isinstance(j, _g.Generator):
                s += str(j.generate())
            else:
                s += str(j)
        return s

    def _next_list(self) -> list:
        s = []
        for i in range(len(self)):
            j = self.get_sequence()[i]
            if isinstance(j, _g.Generator):
                s.append(j.generate())
            else:
                s.append(j)
        return s

    def _next(self) -> Union[str, tuple, list]:
        r = self._next_str() if self._t == str else self._next_list()
        if self._t == tuple:
            r = tuple(r)
        return r

    def get_sequence(self) -> tuple:
        return self._sequential

    def output(self) -> _g.GeneratorOutput:
        r = []
        for _ in range(self._n):
            r.append(self._next())
        return _g.GeneratorOutput(r)

    def save(self) -> None:
        pass



from functools import singledispatch as _singledispatch
from typing import Any, Union
import pickle as _pickle
import abc as _abc
import os as _os


from coota import generator as _g, generator_sequence as _gs


class Operator(object):
    def __init__(self, path: Union[str, _os.PathLike[str]], content: bytes = None):
        self._content: bytes = content
        self._path: path = path

    def e(self) -> bytes:
        i = self.get_id()
        if not -1 < i < 10000:
            raise ValueError("ID must be between 0 and 9999(including).")
        return b"0" * (3 - i // 10) + str(i).encode()

    def get_content(self) -> bytes:
        return self._content

    def get_path(self) -> Union[str, _os.PathLike[str]]:
        return self._path

    @_abc.abstractmethod
    def get_id(self) -> int: pass

    def save(self) -> None:
        class_id = self.e()
        with open(self.get_path(), "wb") as f:
            f.write(class_id + self.get_content())

    @staticmethod
    @_abc.abstractmethod
    def loads(content: bytes) -> Any: pass

    def load(self) -> Any:
        with open(self.get_path(), "rb") as f:
            c = f.read()
            if c[:4] != self.e():
                raise AttributeError(f"File is not formatted as class id {self.e()}.")
            self._content = c[4:]
        return self.loads(content=self.get_content())


class ObjectOperator(Operator):
    def __init__(self, path: Union[str, _os.PathLike[str]], obj: Any = None):
        super().__init__(path, None if obj is None else _pickle.dumps(obj))

    @_abc.abstractmethod
    def get_id(self) -> int:
        pass

    @staticmethod
    def loads(content: bytes) -> Any:
        return _pickle.loads(content)


class StringOperator(Operator):
    @staticmethod
    def loads(content: bytes) -> Any:
        return str(content)

    def get_id(self) -> int:
        return 0


class ChooserOperator(ObjectOperator):
    def get_id(self) -> int:
        return 1


class DistributionOperator(ObjectOperator):
    def get_id(self) -> int:
        return 2


class AssociationOperator(ObjectOperator):
    def get_id(self) -> int:
        return 3


class GeneratorOperator(ObjectOperator):
    def get_id(self) -> int:
        return 4


class GeneratorOutputOperator(ObjectOperator):
    def get_id(self) -> int:
        return 5


class GeneratorSequenceOperator(ObjectOperator):
    def get_id(self) -> int:
        return 6


def load(path: Union[str, _os.PathLike[str]]) -> Any:
    with open(path, "rb") as f:
        class_id = f.read(4)
        if class_id == b"0000":
            return StringOperator(path).load()
        elif class_id == b"0001":
            return ChooserOperator(path).load()
        elif class_id == b"0002":
            return DistributionOperator(path).load()
        elif class_id == b"0003":
            return AssociationOperator(path).load()
        elif class_id == b"0004":
            return GeneratorOperator(path).load()
        elif class_id == b"0005":
            return GeneratorOutputOperator(path).load()
        elif class_id == b"0006":
            return GeneratorSequenceOperator(path).load()
        else:
            raise AttributeError(f"File {path} cannot be loaded because of its unknown type.")


@_singledispatch
def save(obj: Any, path: Union[str, _os.PathLike[str]]) -> None:
    raise TypeError(f"No known case for type {type(obj)}, {type(path)}.")


@save.register(str)
def _(obj: str, path: Union[str, _os.PathLike[str]]) -> None:
    StringOperator(path, obj.encode()).save()


@save.register(_g.Chooser)
def _(obj: _g.Chooser, path: Union[str, _os.PathLike[str]]) -> None:
    ChooserOperator(path, obj).save()


@save.register(_g.Association)
def _(obj: _g.Association, path: Union[str, _os.PathLike[str]]) -> None:
    AssociationOperator(path, obj).save()


@save.register(_g.Generator)
def _(obj: _g.Generator, path: Union[str, _os.PathLike[str]]) -> None:
    GeneratorOperator(path, obj).save()


@save.register(_g.GeneratorOutput)
def _(obj: _g.GeneratorOutput, path: Union[str, _os.PathLike[str]]) -> None:
    GeneratorOutputOperator(path, obj).save()


@save.register(_gs.GeneratorSequence)
def _(obj: _gs.GeneratorSequence, path: Union[str, _os.PathLike[str]]) -> None:
    GeneratorSequenceOperator(path, obj).save()



import setuptools

setuptools.setup(
    name="coota",
    version="1.0.0b9",
    author="ATATC",
    author_email="futerry@outlook.com",
    description="A powerful data-generating library.",
    license='Apache License 2.0',
    long_description="**COOTA**(Come Out Of Thin Air) is a powerful data-generating python library. "
                     "By supporting generator nesting, it allows you to generate a variety of data "
                     "that shows great randomness. It also supports making generators conform to a "
                     "certain distribution and associating two generators.",
    long_description_content_type="text/markdown",
    url="https://github.com/atatc/coota",
    packages=setuptools.find_packages(),
    install_requires=["numpy", "matplotlib", "scipy"],
)
